use std::ops::{Add, Sub, Mul, Div, Rem, Neg};
use std::ops::{AddAssign, SubAssign, MulAssign, DivAssign, RemAssign};

use crate::constants::Constants;
use crate::real::Real;
use crate::trig::radian::Radian;



/// A degree usually denoted by ° (the degree symbol),
/// is a measurement of a plane angle, defined so that a
/// full rotation is 360 degree.
#[derive(Clone, Copy, Eq, Ord, PartialEq, PartialOrd, Hash)]
pub struct Degree<T>
{
   value: T
}



impl<T> Degree<T> where T: Real 
{
   /// Create a new Degree structure with the given value.
   pub fn new(val: T) -> Self
   {
      Degree
      {
         value: val
      }
   }
}

impl<T> ::std::ops::Deref for Degree<T> where T: Real 
{
   type Target = T;

   fn deref(&self) -> &T
   {
      &self.value
   }
}

impl<T> ::std::ops::DerefMut for Degree<T> where T: Real 
{
   fn deref_mut<'a>(&'a mut self) -> &'a mut T
   {
      &mut self.value
   }
}

binary_operator_impl!(Add::add, Real, Degree {value});
binary_operator_impl!(Sub::sub, Real, Degree {value});
binary_operator_impl!(Mul::mul, Real, Degree {value});
binary_operator_impl!(Div::div, Real, Degree {value});
binary_operator_impl!(Rem::rem, Real, Degree {value});

impl<T> ::std::ops::Neg for Degree<T>
   where T: Neg<Output = T>
{
   type Output = Degree<T>;

   fn neg(self) -> Degree<T>
   {
      let mut degree: Degree<T>;

      degree = self;
      degree.value = degree.value.neg();
      degree
   }
}

binary_operator_assign_impl!(AddAssign::add_assign,
                             Real, Degree {value});
binary_operator_assign_impl!(SubAssign::sub_assign,
                             Real, Degree {value});
binary_operator_assign_impl!(MulAssign::mul_assign,
                             Real, Degree {value});
binary_operator_assign_impl!(DivAssign::div_assign,
                             Real, Degree {value});
binary_operator_assign_impl!(RemAssign::rem_assign,
                             Real, Degree {value});

impl ::std::convert::From<Degree<f32>> for f32
{
   fn from(degree: Degree<f32>) -> f32
   {
      degree.value
   }
}

impl ::std::convert::From<Degree<f64>> for f64
{
   fn from(degree: Degree<f64>) -> f64
   {
      degree.value
   }
}

impl<T> ::std::convert::From<T> for Degree<T> where T: Real 
{
   fn from(val: T) -> Degree<T>
   {
      Degree::new(val)
   }
}

impl<T> ::std::convert::From<Radian<T>> for Degree<T> where T: Real 
{
   fn from(radians: Radian<T>) -> Degree<T>
   {
      let degs: T;

      degs = *radians * Constants::INVERSE_PI_DIVIDED_BY_180;
      Degree::new(degs)
   }
}

impl<T> ::std::fmt::Debug for Degree<T>
   where T: Real 
{
   fn fmt(&self, f: &mut ::std::fmt::Formatter) -> ::std::fmt::Result
   {
      ::std::fmt::Display::fmt(self, f)
   }
}

impl<T> ::std::fmt::Display for Degree<T>
   where T: Real 
{
   fn fmt(&self, f: &mut ::std::fmt::Formatter) -> ::std::fmt::Result
   {
      write!(f, "Degree({})", self.value)
   }
}
