//! This module defines 2, 3, and 4 component [Vector][1] structures.
//!
//! [1]: https://en.wikipedia.org/wiki/Vector_(mathematics_and_physics)
use std::fmt::{Error, Formatter, Debug, Display};
use std::ops::{Add, Sub, Mul, Div, Rem, Neg};
use std::ops::{AddAssign, SubAssign, MulAssign, DivAssign, RemAssign};

use crate::zero::Zero;
use crate::one::One;
use crate::number::Number;
use crate::real::Real;
use crate::trig::Radian;
use crate::trig::Trig;



/// A trait that defines the minimum set of
/// functions a [Vector][1] must implement.
///
/// [1]: https://en.wikipedia.org/wiki/Vector_(mathematics_and_physics)
pub trait Vector<T>: Debug + Display + Clone + Default + Zero + One
   where T: Number
{
   // Creation functions.
   /// Create a Vector from a single value.
   ///
   ///```
   /// // Create a new Vector3<f32> where all
   /// // the components are 3.4.
   /// use sigils::vector::*;
   ///
   /// let vector = Vector3::<f32>::from_value(3.4f32);
   /// # assert_eq!(vector.x, 3.4f32);
   /// # assert_eq!(vector.y, 3.4f32);
   /// # assert_eq!(vector.z, 3.4f32);
   ///```
   fn from_value(val: T) -> Self;

   /// Create an identity Vector. All components will be set to one.
   ///
   /// ```
   /// // Create a new Vector2<i32> where all
   /// // the components are 1.
   /// use sigils::vector::*;
   ///
   /// let vector = Vector2::<i32>::identity();
   /// # assert_eq!(vector.x, 1i32);
   /// # assert_eq!(vector.y, 1i32);
   /// ```
   fn identity() -> Self
   {
      Self::from_value(T::one())
   }


   // Property functions.
   /// Returns the size, how many components the Vector
   /// has, of this Vector.
   fn get_size(&self) -> u8;


   // Scalar operations that change this Vector.
   /// Add a scalar value to this Vector.
   ///
   /// ```
   /// // Create a new Vector2<i32> where all
   /// // the components are 1. Then add 1 to it.
   /// use sigils::vector::*;
   ///
   /// let mut vector = Vector2::<i32>::identity();
   /// vector.add_scalar(1i32);
   /// # assert_eq!(vector.x, 2i32);
   /// # assert_eq!(vector.y, 2i32);
   /// ```
   fn add_scalar(&mut self, scalar: T);

   /// Subtract a scalar value from this Vector.
   ///
   /// ```
   /// // Create a new Vector4<f64> where all
   /// // the components are 1. Then subtract 1 from it.
   /// use sigils::vector::*;
   ///
   /// let mut vector = Vector4::<f64>::identity();
   /// vector.sub_scalar(1.0f64);
   /// # assert_eq!(vector.x, 0f64);
   /// # assert_eq!(vector.y, 0f64);
   /// # assert_eq!(vector.z, 0f64);
   /// # assert_eq!(vector.w, 0f64);
   /// ```
   fn sub_scalar(&mut self, scalar: T);

   /// Multiply this Vector by a scalar value.
   ///
   /// ```
   /// // Create a new Vector4<f64> where all
   /// // the components are 5. Then multiply it by 5.
   /// use sigils::vector::*;
   ///
   /// let mut vector = Vector4::<f32>::from_value(5f32);
   /// vector.mul_scalar(5f32);
   /// # assert_eq!(vector.x, 25f32);
   /// # assert_eq!(vector.y, 25f32);
   /// # assert_eq!(vector.z, 25f32);
   /// # assert_eq!(vector.w, 25f32);
   /// ```
   fn mul_scalar(&mut self, scalar: T);

   /// Divide this Vector by a scalar value.
   ///
   /// ```
   /// // Create a new Vector3<u8> where all
   /// // the components are 25. Then divide it by 5.
   /// use sigils::vector::*;
   ///
   /// let mut vector = Vector3::<u8>::from_value(25u8);
   /// vector.div_scalar(5u8);
   /// # assert_eq!(vector.x, 5u8);
   /// # assert_eq!(vector.y, 5u8);
   /// # assert_eq!(vector.z, 5u8);
   /// ```
   fn div_scalar(&mut self, scalar: T);

   /// Divide this Vector by a scalar value and store the remainder.
   ///
   /// ```
   /// // Create a new Vector3<isize> where all
   /// // the components are 25. Then divide it by 7
   /// // and take the remainder.
   /// use sigils::vector::*;
   ///
   /// let mut vector = Vector3::<isize>::from_value(25isize);
   /// vector.rem_scalar(7isize);
   /// # assert_eq!(vector.x, 4isize);
   /// # assert_eq!(vector.y, 4isize);
   /// # assert_eq!(vector.z, 4isize);
   /// ```
   fn rem_scalar(&mut self, scalar: T);


   // Vector operations that change this Vector.
   /// Add a Vector to this Vector.
   ///
   /// ```
   /// // Create a new Vector2<i32> where all
   /// // the components are 1. Then add
   /// // another Vector2<i32>, where all the
   /// // components are 1, to it.
   /// use sigils::vector::*;
   ///
   /// let mut vector = Vector2::<i32>::identity();
   /// let vector_two = Vector2::<i32>::identity();
   /// vector.add_vector(&vector_two);
   /// # assert_eq!(vector.x, 2i32);
   /// # assert_eq!(vector.y, 2i32);
   /// ```
   fn add_vector(&mut self, vector: &Self);

   /// Subtract a Vector from this Vector.
   ///
   /// ```
   /// // Create a new Vector3<i64> where all
   /// // the components are 5. Then subtract
   /// // a Vector3<i64>, where all the
   /// // components are 2, from it.
   /// use sigils::vector::*;
   ///
   /// let mut vector = Vector3::<i64>::from_value(5i64);
   /// let vector_two = Vector3::<i64>::from_value(2i64);
   /// vector.sub_vector(&vector_two);
   /// # assert_eq!(vector.x, 3i64);
   /// # assert_eq!(vector.y, 3i64);
   /// # assert_eq!(vector.z, 3i64);
   /// ```
   fn sub_vector(&mut self, vector: &Self);

   /// Multiply this Vector by a Vector.
   ///
   /// ```
   /// // Create a new Vector4<f32> where all
   /// // the components are 6.5. Then multiply
   /// // it by another Vector4<f32>, where all the
   /// // components are 4.0.
   /// use sigils::vector::*;
   ///
   /// let mut vector = Vector4::<f32>::from_value(6.5f32);
   /// let vector_two = Vector4::<f32>::from_value(4.0f32);
   /// vector.mul_vector(&vector_two);
   /// # assert_eq!(vector.x, 26.0f32);
   /// # assert_eq!(vector.y, 26.0f32);
   /// # assert_eq!(vector.z, 26.0f32);
   /// # assert_eq!(vector.w, 26.0f32);
   /// ```
   fn mul_vector(&mut self, vector: &Self);

   /// Divide this Vector by a Vector.
   ///
   /// ```
   /// // Create a new Vector2<f64> where all
   /// // the components are 19. Then
   /// // divide it by another Vector2<f64>,
   /// // where all the components are 2.0.
   /// use sigils::vector::*;
   ///
   /// let mut vector = Vector2::<f64>::from_value(19.0f64);
   /// let vector_two = Vector2::<f64>::from_value(2.0f64);
   /// vector.div_vector(&vector_two);
   /// # assert_eq!(vector.x, 9.5f64);
   /// # assert_eq!(vector.y, 9.5f64);
   /// ```
   fn div_vector(&mut self, vector: &Self);

   /// Divide this Vector by a Vector and store the remainder.
   ///
   /// ```
   /// // Create a new Vector3<usize> where all
   /// // the components are 22. Then
   /// // divide it by another Vector3<usize>,
   /// // where all the components are 6.
   /// use sigils::vector::*;
   ///
   /// let mut vector = Vector3::<usize>::from_value(22usize);
   /// let vector_two = Vector3::<usize>::from_value(6usize);
   /// vector.rem_vector(&vector_two);
   /// # assert_eq!(vector.x, 4usize);
   /// # assert_eq!(vector.y, 4usize);
   /// # assert_eq!(vector.z, 4usize);
   /// ```
   fn rem_vector(&mut self, vector: &Self);


   // Basic Vector functions.
   /// Get the sum of all the components of the Vector.
   ///
   ///```
   /// // Create a new Vector4<f32> where all
   /// // the components are 5.125. Then get
   /// // the sum of the Vector's components.
   /// use sigils::vector::*;
   ///
   /// let vector: Vector4<f32> = Vector4::<f32>::from_value(5.125f32);
   /// let sum: f32 = vector.get_sum();
   /// # assert_eq!(sum, 20.5f32);
   ///```
   fn get_sum(&self) -> T;

   /// Get the product of all the components of the Vector.
   ///
   ///```
   /// // Create a new Vector3<isize> where all
   /// // the components are 3. Then get
   /// // the product of the Vector's components.
   /// use sigils::vector::*;
   ///
   /// let vector: Vector3<isize> = Vector3::<isize>::from_value(3isize);
   /// let sum: isize = vector.get_product();
   /// # assert_eq!(sum, 27isize);
   ///```
   fn get_product(&self) -> T;

   /// Get the dot product between this and another Vector.
   ///
   ///```
   /// // Create a two new Vector3<i64> where all
   /// // the components are 3. Then get the
   /// // dot product of the two Vectors.
   /// use sigils::vector::*;
   /// use sigils::Radian;
   ///
   /// let vector: Vector3<i32> = Vector3::<i32>::from_value(3i32);
   /// let vector_two: Vector3<i32> = Vector3::<i32>::from_value(3i32);
   /// let dotProduct: i32 = vector.dot(&vector_two);
   /// # assert_eq!(dotProduct, 27i32);
   ///```
   fn dot(&self, vector: &Self) -> T;
}

/// Defines the [EuclideanVector][1] trait.
///
/// [1]: https://en.wikipedia.org/wiki/Euclidean_vector
pub trait EuclideanVector<T> : Vector<T> where T: Trig
{
   /// Get the length of the Vector.
   ///
   ///```
   /// // Get the length of the Vector.
   /// use sigils::vector::*;
   ///
   /// let vector = Vector4::<f32>::new(5.5f32, 5.5f32, 5.5f32, 5.5f32);
   /// let length = vector.get_length();
   /// # assert_eq!(length, 11.0f32);
   ///```
   fn get_length(&self) -> T
   {
      let length: T;

      length = self.dot(self).sqrt();
      assert!(length.is_finite());

      length
   }

   /// Get the squared length of the Vector.
   ///
   ///```
   /// // Get the squared length of the Vector.
   /// use sigils::vector::*;
   ///
   /// let vector = Vector4::<f32>::new(5.5f32, 5.5f32, 5.5f32, 5.5f32);
   /// let length = vector.get_length_squared();
   /// # assert_eq!(length, 121.0f32);
   ///```
   fn get_length_squared(&self) -> T
   {
      self.dot(self)
   }

   /// Normalizes the Vector by multplying all the
   /// components by 1/|V|.
   ///
   ///```
   /// // Creae a Vector and normalize it.
   /// use sigils::vector::*;
   ///
   /// let vector = Vector4::<f32>::new(5.0f32, 2.5f32, 5.5f32, 5.5f32);
   /// let normal = vector.normalize();
   /// # assert_eq!(normal.x, 0.5219958f32);
   /// # assert_eq!(normal.y, 0.2609979f32);
   /// # assert_eq!(normal.z, 0.5741953f32);
   /// # assert_eq!(normal.w, 0.5741953f32);
   ///```
   fn normalize(&self) -> Self
   {
      let mut new_vector: Self;

      new_vector = self.clone();
      new_vector.mul_scalar(self.get_length().recip());
      new_vector
   }

   /// Normalizes the Vector by multplying all the
   /// components by 1/|V|.
   ///
   ///```
   /// // Creae a Vector and normalize it.
   /// use sigils::vector::*;
   ///
   /// let mut vector = Vector4::<f32>::new(5.0f32, 2.5f32, 5.5f32, 5.5f32);
   /// vector.normalize_self();
   /// # assert_eq!(vector.x, 0.5219958f32);
   /// # assert_eq!(vector.y, 0.2609979f32);
   /// # assert_eq!(vector.z, 0.5741953f32);
   /// # assert_eq!(vector.w, 0.5741953f32);
   ///```
   fn normalize_self(&mut self)
   {
      let length: T;

      length = self.get_length().recip();
      self.mul_scalar(length)
   }

   /// Determines if this Vector is perpendicular to another Vector.
   // TODO: Add an example here.
   fn is_perpendicular_to(&self, vector: &Self) -> bool
   {
      // TODO: Make this work with a fudge factor since floats
      //       are tricky.
      self.dot(vector) == T::zero()
   }

   /// Calculates the angle between this vector and
   /// another Vector, in radians.
   // TODO: Add an example here.
   fn angle(&self, vector: &Self) -> Radian<T>;

   /// Linearly interpolate the length of this Vector
   /// towards the length of another Vector by a given amount.
   // TODO: Add an example here.
   fn lerp(&self, vector: &Self, amount: T) -> Self
   {
      let mut temp: Self;
      let mut new_vector: Self;

      temp = vector.clone();
      temp.sub_vector(&self);
      temp.mul_scalar(amount);

      new_vector = self.clone();
      new_vector.add_vector(&temp);
      new_vector
   }

   /// Linearly interpolate the length of this Vector
   /// towards the length of another Vector by a given amount.
   // TODO: Add an example here.
   fn lerp_self(&mut self, vector: &Self, amount: T)
   {
      let mut temp: Self;

      temp = vector.clone();
      temp.mul_scalar(amount);
      self.add_vector(&temp);
   }
}


// Create a few macros to define the three Vector
// structs: Vector2, Vector3, and Vector4.
/// A macro that will help with adding and
/// multiplying components of a vector.
macro_rules! perform_method_on_components
{
   ($method: ident, {$x: expr, $y: expr}) =>
   {
      $x.$method($y)
   };

   ($method: ident, {$x: expr, $y: expr, $z: expr}) =>
   {
      $x.$method($y).$method($z)
   };

   ($method: ident, {$x: expr, $y: expr, $z: expr, $w: expr}) =>
   {
      $x.$method($y).$method($z).$method($w)
   };
}

/// A macro that defines a Vector structure
/// that implements the Vector trait.
macro_rules! define_vector
{
   ($structName: ident <$T: ident> {$($field: ident),+}, $field_count: expr) =>
   {
      // Do not add a where clause here or implementing
      // generic traits were the generic is different
      // will not work properly.
      /// This structure was defined with the macro
      /// define_vector.
      ///
      /// Please excuse the lack of comments and refer
      /// to the trait definitions for comments about
      /// the implementation.
      #[derive(PartialEq, Eq, Copy, Clone, Hash)]
      pub struct $structName<T>
      {
         $(pub $field: T),+
      }

      impl<$T> $structName<$T> where T: Number
      {
         pub fn new($($field: $T),+) -> $structName<$T>
         {
            $structName {$($field: $field),+}
         }
      }

      impl<T> Vector<T> for $structName<T> where T: Number
      {
         fn from_value(val: T) -> $structName<T>
         {
            $structName {$($field: val),+}
         }

         fn get_size(&self) -> u8
         {
            $field_count
         }

         fn add_scalar(&mut self, scalar: T)
         {
            *self = &*self + scalar;
         }

         fn sub_scalar(&mut self, scalar: T)
         {
            *self = &*self - scalar;
         }

         fn mul_scalar(&mut self, scalar: T)
         {
            *self = &*self * scalar;
         }

         fn div_scalar(&mut self, scalar: T)
         {
            *self = &*self / scalar;
         }

         fn rem_scalar(&mut self, scalar: T)
         {
            *self = &*self % scalar;
         }


         fn add_vector(&mut self, vector: &$structName<T>)
         {
            *self = &*self + vector;
         }

         fn sub_vector(&mut self, vector: &$structName<T>)
         {
            *self = &*self - vector;
         }

         fn mul_vector(&mut self, vector: &$structName<T>)
         {
            *self = &*self * vector;
         }

         fn div_vector(&mut self, vector: &$structName<T>)
         {
            *self = &*self / vector;
         }

         fn rem_vector(&mut self, vector: &$structName<T>)
         {
            *self = &*self % vector;
         }


         fn get_sum(&self) -> T
         {
            perform_method_on_components!(add, {$(self.$field),+})
         }

         fn get_product(&self) -> T
         {
            perform_method_on_components!(mul, {$(self.$field),+})
         }

         fn dot(&self, vector: &$structName<T>) -> T
         {
            self.mul(vector).get_sum()
         }
      }

      // Implement the Debug trait for the Vector.
      impl<T> Debug for $structName<T> where T: Number
      {
         #[allow(unused_assignments)]
         fn fmt(&self, formatter: &mut Formatter) -> Result<(), Error>
         {
            let mut count: u8;

            // Use the size of the Vector to
            // pretty up the printing.
            count = self.get_size();

            write!(formatter, "<")?;
            $(
               write!(formatter, "{:?}", self.$field)?;
               if count > 0
               {
                  write!(formatter, ", ")?;
               }
               count -= 1;
            )*
            write!(formatter, ">")
         }
      }

      // Implement the Display trait for the Vector.
      impl<T> Display for $structName<T> where T: Number
      {
         #[allow(unused_assignments)]
         fn fmt(&self, formatter: &mut Formatter) -> Result<(), Error>
         {
            let mut count: u8;

            // Use the size of the Vector to
            // pretty up the printing.
            count = self.get_size();

            write!(formatter, "<")?;
            $(
               write!(formatter, "{}", self.$field)?;
               if count > 1
               {
                  write!(formatter, ", ")?;
               }
               count -= 1;
            )*
            write!(formatter, ">")
         }
      }

      // Implement the Default trait for the Vector.
      impl<T> Default for $structName<T> where T: Number
      {
         fn default() -> $structName<T>
         {
            $structName::identity()
         }
      }

      // Implement the Zero and One traits for the Vector.
      impl<T> Zero for $structName<T> where T: Number
      {
         fn zero() -> $structName<T>
         {
            $structName::from_value(T::zero())
         }
      }

      impl<T> One for $structName<T> where T: Number
      {
         fn one() -> $structName<T>
         {
            $structName::identity()
         }
      }

      // Implement the negation operator for the Vector structure.
      impl<T> Neg for $structName<T> where T: Neg<Output = T>
      {
         type Output = $structName<T>;

         fn neg(self) -> $structName<T>
         {
            $structName::<T> {$($field: -self.$field),+}
         }
      }

      // Implement the binary operations for this Vector structure.
      binary_operator_impl!(Add::add, Number, $structName {$($field),+});
      binary_operator_impl!(Sub::sub, Number, $structName {$($field),+});
      binary_operator_impl!(Mul::mul, Number, $structName {$($field),+});
      binary_operator_impl!(Div::div, Number, $structName {$($field),+});
      binary_operator_impl!(Rem::rem, Number, $structName {$($field),+});

      // Handle the assignment operators.
      binary_operator_assign_impl!(AddAssign::add_assign,
                                   Number, $structName {$($field),+});
      binary_operator_assign_impl!(SubAssign::sub_assign,
                                   Number, $structName {$($field),+});
      binary_operator_assign_impl!(MulAssign::mul_assign,
                                   Number, $structName {$($field),+});
      binary_operator_assign_impl!(DivAssign::div_assign,
                                   Number, $structName {$($field),+});
      binary_operator_assign_impl!(RemAssign::rem_assign,
                                   Number, $structName {$($field),+});
   }
}


// Define the Vector2, Vector3, and Vector4 structures.
define_vector!(Vector2<T> {x, y}, 2u8);
define_vector!(Vector3<T> {x, y, z}, 3u8);
define_vector!(Vector4<T> {x, y, z, w}, 4u8);



// Implements operations specific to the different
// Vector structure sizes.
impl<T> Vector2<T> where T: Number
{
   /// Create a unit Vector in the X direction.
   ///
   ///```
   /// // Create a new X unit Vector.
   /// use sigils::vector::*;
   ///
   /// let vector = Vector2::<f64>::unit_x();
   /// # assert_eq!(vector.x, 1.0f64);
   /// # assert_eq!(vector.y, 0.0f64);
   ///```
   pub fn unit_x() -> Vector2<T>
   {
      Vector2::new(T::one(), T::zero())
   }

   /// Create a unit Vector in the Y direction.
   ///
   ///```
   /// // Create a new Y unit Vector.
   /// use sigils::vector::*;
   ///
   /// let vector = Vector2::<f64>::unit_y();
   /// # assert_eq!(vector.x, 0.0f64);
   /// # assert_eq!(vector.y, 1.0f64);
   ///```
   pub fn unit_y() -> Vector2<T>
   {
      Vector2::new(T::zero(), T::one())
   }

   /// Calculate the perpendicular dot product.
   // TODO: Add an example.
   pub fn perpendicular_dot(&self, vector: &Vector2<T>) -> T
   {
      (self.x * vector.y) - (self.y * vector.x)
   }

   /// Adds a z component with the given value to the Vector.
   ///
   ///```
   /// // Create a new Vector2 X unit Vector and
   /// // extend it to a Vector3.
   /// use sigils::vector::*;
   ///
   /// let vector = Vector2::<f32>::unit_x();
   /// let biggerVec = vector.extend(6.5f32);
   /// # assert_eq!(biggerVec.x, 1.0f32);
   /// # assert_eq!(biggerVec.y, 0.0f32);
   /// # assert_eq!(biggerVec.z, 6.5f32);
   ///```
   pub fn extend(&self, val: T) -> Vector3<T>
   {
      Vector3::new(self.x, self.y, val)
   }
}

impl<T> Vector3<T> where T: Number
{
   /// Create a unit Vector in the X direction.
   ///
   ///```
   /// // Create a new X unit Vector.
   /// use sigils::vector::*;
   ///
   /// let vector = Vector3::<f64>::unit_x();
   /// # assert_eq!(vector.x, 1.0f64);
   /// # assert_eq!(vector.y, 0.0f64);
   /// # assert_eq!(vector.z, 0.0f64);
   ///```
   pub fn unit_x() -> Vector3<T>
   {
      Vector3::new(T::one(), T::zero(), T::zero())
   }

   /// Create a unit Vector in the Y direction.
   ///
   ///```
   /// // Create a new Y unit Vector.
   /// use sigils::vector::*;
   ///
   /// let vector = Vector3::<f64>::unit_y();
   /// # assert_eq!(vector.x, 0.0f64);
   /// # assert_eq!(vector.y, 1.0f64);
   /// # assert_eq!(vector.z, 0.0f64);
   ///```
   pub fn unit_y() -> Vector3<T>
   {
      Vector3::new(T::zero(), T::one(), T::zero())
   }

   /// Create a unit Vector in the Z direction.
   ///
   ///```
   /// // Create a new Z unit Vector.
   /// use sigils::vector::*;
   ///
   /// let vector = Vector3::<f64>::unit_z();
   /// # assert_eq!(vector.x, 0.0f64);
   /// # assert_eq!(vector.y, 0.0f64);
   /// # assert_eq!(vector.z, 1.0f64);
   ///```
   pub fn unit_z() -> Vector3<T>
   {
      Vector3::new(T::zero(), T::zero(), T::one())
   }

   /// Calculate the cross product between this and another Vector.
   /// It will return a Vector perpendicular to both of the input Vectors.
   pub fn cross(&self, vector: &Vector3<T>) -> Vector3<T>
   {
      Vector3::new((self.y * vector.z) - (self.z * vector.y),
                   (self.z * vector.x) - (self.x * vector.z),
                   (self.x * vector.y) - (self.y * vector.x))
   }

   /// Calculate the cross product between this and another Vector
   /// and store it in this Vector.
   /// The final result will be a Vector perpendicular to both
   /// of the input Vectors.
   pub fn cross_self(&mut self, vector: &Vector3<T>)
   {
      *self = self.cross(&vector);
   }

   /// Adds a w component with the given value to the Vector.
   ///
   ///```
   /// // Create a new Vector3 X unit Vector and
   /// // extend it to a Vector4.
   /// use sigils::vector::*;
   ///
   /// let vector = Vector3::<f32>::unit_x();
   /// let biggerVec = vector.extend(6.5f32);
   /// # assert_eq!(biggerVec.x, 1.0f32);
   /// # assert_eq!(biggerVec.y, 0.0f32);
   /// # assert_eq!(biggerVec.z, 0.0f32);
   /// # assert_eq!(biggerVec.w, 6.5f32);
   ///```
   pub fn extend(&self, val: T) -> Vector4<T>
   {
      Vector4::new(self.x, self.y, self.z, val)
   }

   /// Discards the z component of the Vector.
   pub fn truncate(&self) -> Vector2<T>
   {
      self.truncate_component(2)
   }

   /// Discards the component at the given index from the Vector.
   pub fn truncate_component(&self, index: isize) -> Vector2<T>
   {
      match index
      {
         0 => Vector2::new(self.y, self.z),
         1 => Vector2::new(self.x, self.z),
         2 => Vector2::new(self.x, self.y),
         _ => panic!("Component index {:?} is out of range.", index)
      }
   }
}

impl<T> Vector4<T> where T: Number
{
   /// Create a unit Vector in the X direction.
   ///
   ///```
   /// // Create a new X unit Vector.
   /// use sigils::vector::*;
   ///
   /// let vector = Vector4::<f64>::unit_x();
   /// # assert_eq!(vector.x, 1.0f64);
   /// # assert_eq!(vector.y, 0.0f64);
   /// # assert_eq!(vector.z, 0.0f64);
   /// # assert_eq!(vector.w, 0.0f64);
   ///```
   pub fn unit_x() -> Vector4<T>
   {
      Vector4::new(T::one(), T::zero(), T::zero(), T::zero())
   }

   /// Create a unit Vector in the Y direction.
   ///
   ///```
   /// // Create a new Y unit Vector.
   /// use sigils::vector::*;
   ///
   /// let vector = Vector4::<f64>::unit_y();
   /// # assert_eq!(vector.x, 0.0f64);
   /// # assert_eq!(vector.y, 1.0f64);
   /// # assert_eq!(vector.z, 0.0f64);
   /// # assert_eq!(vector.w, 0.0f64);
   ///```
   pub fn unit_y() -> Vector4<T>
   {
      Vector4::new(T::zero(), T::one(), T::zero(), T::zero())
   }

   /// Create a unit Vector in the Z direction.
   ///
   ///```
   /// // Create a new Z unit Vector.
   /// use sigils::vector::*;
   ///
   /// let vector = Vector4::<f64>::unit_z();
   /// # assert_eq!(vector.x, 0.0f64);
   /// # assert_eq!(vector.y, 0.0f64);
   /// # assert_eq!(vector.z, 1.0f64);
   /// # assert_eq!(vector.w, 0.0f64);
   ///```
   pub fn unit_z() -> Vector4<T>
   {
      Vector4::new(T::zero(), T::zero(), T::one(), T::zero())
   }

   /// Create a unit Vector in the W direction.
   ///
   ///```
   /// // Create a new W unit Vector.
   /// use sigils::vector::*;
   ///
   /// let vector = Vector4::<f64>::unit_w();
   /// # assert_eq!(vector.x, 0.0f64);
   /// # assert_eq!(vector.y, 0.0f64);
   /// # assert_eq!(vector.z, 0.0f64);
   /// # assert_eq!(vector.w, 1.0f64);
   ///```
   pub fn unit_w() -> Vector4<T>
   {
      Vector4::new(T::zero(), T::zero(), T::zero(), T::one())
   }

   /// Discards the w component of the Vector.
   pub fn truncate(&self) -> Vector3<T>
   {
      self.truncate_component(3)
   }

   /// Discards the component at the given index from the Vector.
   pub fn truncate_component(&self, index: isize) -> Vector3<T>
   {
      match index
      {
         0 => Vector3::new(self.y, self.z, self.w),
         1 => Vector3::new(self.x, self.z, self.w),
         2 => Vector3::new(self.x, self.y, self.w),
         3 => Vector3::new(self.x, self.y, self.z),
         _ => panic!("Component index {:?} is out of range.", index)
      }
   }
}

// Implement the angle specific portion of the EuclideanVector.
impl<T> EuclideanVector<T> for Vector2<T> where T: Trig
{
   fn angle(&self, vector: &Vector2<T>) -> Radian<T>
   {
      // Shortcut.
      Trig::atan2(self.perpendicular_dot(vector), self.dot(vector))
   }
}

impl<T> EuclideanVector<T> for Vector3<T> where T: Trig
{
   fn angle(&self, vector: &Vector3<T>) -> Radian<T>
   {
      // Shortcut.
      Trig::atan2(self.cross(vector).get_length(), self.dot(vector))
   }
}

impl<T> EuclideanVector<T> for Vector4<T> where T: Trig
{
   fn angle(&self, vector: &Vector4<T>) -> Radian<T>
   {
      // The basic equation, this would work for any size Vector.
      Trig::acos(self.dot(vector) / (self.get_length()*vector.get_length()))
   }
}
